
% RR-Rotation
avl_rotation(node(X,2,L,node(Y,YBalance,RL,RR)),
              node(Y,BR,node(X,BL,L,RL),RR) ) :-
   YBalance=  1  -> BL =  0, BR =  0 ;
   YBalance=  0  -> BL =  1, BR = -1.
% RL-Rotation
avl_rotation( node(X,2,L,node(Y,-1,node(Z,ZBalance,RLL,RLR),RR)),
              node(Z,0,node(X,BL,L,RLL),node(Y,BR,RLR,RR)) ):-
   ZBalance=  1  -> BL = -1, BR =  0 ;
   ZBalance=(-1) -> BL =  0, BR =  1.
% LL-Rotation
avl_rotation( node(X,-2,node(Y,YBalance,LL,LR),R),
              node(Y,BL,LL,node(X,BR,LR,R)) ):-
   YBalance=(-1) -> BL =  0, BR =  0 ;
   YBalance=  0  -> BL =  1, BR = -1.
% LR-Rotation
avl_rotation( node(X,-2,node(Y,1,LL,node(Z,ZBalance,LRL, LRR)), R),
              node(Z,0,node(Y,BL,LL,LRL), node(X,BR,LRR,R)) ):-
   ZBalance=  1  -> BL =  1, BR=  0 ;
   ZBalance=(-1) -> BL =  0, BR= -1.


avl_insert(Q, nil, node(Q, 0, nil,nil), 1).
avl_insert(Q, node(X, Balance, L, R), Tree, HeightChanged):-
   (X < Q
   -> (avl_insert(Q, R, NewR, RekHeightChanged),
      NewL = L,
      NewBalance is Balance + RekHeightChanged)
   ;  (avl_insert(Q, L, NewL, RekHeightChanged),
      NewR = R,
      NewBalance is Balance - RekHeightChanged)
   ),
   (avl_rotation(node(X,NewBalance,NewL,NewR), Tree)
   -> HeightChanged = 0
   ;  (Tree=node(X,NewBalance,NewL,NewR),
      ((RekHeightChanged=0 ; NewBalance=0)
      -> HeightChanged = 0
      ;  HeightChanged = 1)
      )
   ).


avl_fix(node(X,NewBalance,L,R),Fixed,RekHeightChanged,HeightChanged):-
   avl_rotation(node(X,NewBalance,L,R),Fixed)
   -> HeightChanged = 0
   ;  (Fixed=node(X,NewBalance,L,R),
      ((RekHeightChanged=0 ; NewBalance\=0)
      -> HeightChanged = 0
      ;  HeightChanged = 1)
   ).

avl_get_leftmost(node(X,_,nil,R), X, R, 1).
avl_get_leftmost(node(X,Balance,L,R), Leftmost, Tree, HeightChanged):-
   avl_get_leftmost(L, Leftmost, LRest, RekHeightChanged),
   NewBalance is Balance + RekHeightChanged,
   avl_fix(node(X,NewBalance,LRest,R),Tree, RekHeightChanged, HeightChanged).

avl_delete(Q, node(Q, _, L, nil), L, 1).
avl_delete(Q, node(Q, Balance, L, R), Tree, HeightChanged) :-
   avl_get_leftmost(R, Leftmost, RRest, RekHeightChanged),
   NewBalance is Balance - RekHeightChanged,
   avl_fix(node(Leftmost, NewBalance, L, RRest), Tree, RekHeightChanged, HeightChanged).

avl_delete(Q, node(X, Balance, L, R), Tree, HeightChanged):-
   X < Q,
   avl_delete(Q,R,NewR,RekHeightChanged),
   NewBalance is Balance - RekHeightChanged,
   avl_fix(node(X,NewBalance,L,NewR), Tree, RekHeightChanged, HeightChanged).
avl_delete(Q, node(X, Balance, L, R), Tree, HeightChanged):-
   Q =< X,
   avl_delete(Q,L,NewL,RekHeightChanged),
   NewBalance is Balance + RekHeightChanged,
   avl_fix(node(X,NewBalance,NewL,R), Tree, RekHeightChanged, HeightChanged).


avl_insert_list([], Tree, Tree).
avl_insert_list([X|R], Tree, Result):-
   avl_insert(X,Tree,NewTree,_),
   avl_insert_list(R,NewTree,Result).

avl_delete_list([], Tree, Tree).
avl_delete_list([X|R], Tree, Result):-
   avl_delete(X,Tree,NewTree,_),
   avl_delete_list(R,NewTree,Result).